﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;
using CefSharp;
using CefSharp.Wpf;
using edu.stanford.nlp.ling;
using edu.stanford.nlp.pipeline;
using edu.stanford.nlp.util;
using ikvm.extensions;
using java.util;
using ShareReadability.Container;
using Color = System.Drawing.Color;

namespace ShareReadability.ReadabilityAnalysis
{
    /// @file ReadabilityAnalysisPipeline.cs
    /// <summary>
    /// This class represents the readability analysis pipeline.
    /// </summary>
    public class ReadabilityAnalysisPipeline
    {
        // Path to the folder with models extracted from `stanford-corenlp-3.9.1-models.jar`
        private const string JarRoot = @"./data/stanford-corenlp-3.9.1-models/";

        // Javascript to click the recompile button
        private const string Script = @"document.getElementsByClassName(""btn btn-recompile"")[0].click();";

        // LaTeX packages to inject temporary before the document tag to support highlighting
        private const string Package = "\\usepackage[rgb]{xcolor}\r\n\\usepackage[author={ShareReadability}]{pdfcomment}\r\n";
        private const string Document = "\\begin{document}";

        // Same alpha value as for LinearGradientBrush in MainWindow.xaml (191 dec = BF hex)
        private const int Alpha = 191;

        // Background color of the pdf
        private static readonly Color BackgroundColor = Color.White;

        // The maximum words per sentence, everything above is very difficult to read
        private const int MaxSentenceLength = 30;


        // Patterns
        private const string PatternFilter = @"(%(.+))|" + // comments '% foo bar'
                                             @"(\\\w+(\[.+\])?(({.+}){1,3})?(\[.+\])?)|" + // tags '\foo', '\foo{bar}', '\foo[42]{bar}', '\foo{bar}[42]', '\foo[13]{bar}[37]' 
                                             @"((.+)=(.+))(,?)|" + // 'foo=bar'
                                             @"(\\\\)|";// + // '\\'
        // Extract plain text
        private const string PatternExtractPlaintext = @"\\\w+{(.+)}";
        // MIND: Using this pattern because parsing the text to generate the phrase structure tree for each sentence using NLP is computationally to expensive
        private const string PatterSentenceStructureComplexity = @"(\swho\s|\swhich\s|\swhenever\s|\safter\s|\sif\s|\swhen\s|\sbut\s|\sor\s|\sbecause\s|\sand\s|\sthat\s|\swhose\s|,|\sneither\s|\s-\s|\(|\)|\[|\]|\{|\})";


        // The part of speech labels
        // https://stackoverflow.com/questions/1833252/java-stanford-nlp-part-of-speech-labels
        private static readonly HashSet<string> Verbs = new HashSet<string>(new[] { "VB", "VBD", "VBG", "VBN", "VBP", "VBZ" });
        private static readonly HashSet<string> Nouns = new HashSet<string>(new[] { "NN", "NNS", "NNP", "NNPS" });

        // Some variables from the GUI
        public static Settings Settings { get; set; }
        public static MainWindow MainWindow { get; set; }

        // Some variables for the pipeline
        private static List<string> _whitelist;
        private static StanfordCoreNLP _pipeline;

        /// <summary>
        /// Initializes the ReadabilityAnalysisPipeline.
        /// </summary>
        public static void Initialize()
        {
            // Set decimal seperator from ',' to '.'
            CultureInfo customCulture = (CultureInfo)Thread.CurrentThread.CurrentCulture.Clone();
            customCulture.NumberFormat.NumberDecimalSeparator = ".";
            Thread.CurrentThread.CurrentCulture = customCulture;

            // Annotation pipeline configuration
            var props = new java.util.Properties();
            props.setProperty("annotators", "tokenize, ssplit, pos, lemma");

            // We should change current directory, so StanfordCoreNLP could find all the model files automatically
            var curDir = Environment.CurrentDirectory;
            Directory.SetCurrentDirectory(JarRoot);
            _pipeline = new StanfordCoreNLP(props);
            Directory.SetCurrentDirectory(curDir);

            _whitelist = new List<string>(File.ReadAllLines(Settings.WhitelistPath));
        }

        /// <summary>
        /// Executes the ReadabilityAnalysisPipeline.
        /// </summary>
        /// <param name="chromiumWebBrowser">The browser to interact with.</param>
        public static void ExecutePipeline(ChromiumWebBrowser chromiumWebBrowser)
        {
            // Disable Main Menu
            MainWindow.EnableMainMenu(false);

            // Try to process pipeline
            try
            {
                // 1. Get Raw text
                var rawText = CopyToClipboard(chromiumWebBrowser);

                // 2. Filter text
                var filteredText = GetFilteredText(rawText);

                // 3. Extract sentences
                var sentences = ExtractSentencesFromText(filteredText);

                // 4. Analyze the sentences
                AnalyzeSentences(sentences);

                // 5. Highlight the sentences
                var highlightedText = GetHighlightedText(sentences, rawText);

                // 6. Paste the modified text and recompile
                PasteFromClipboard(chromiumWebBrowser, highlightedText);
                Thread.Sleep(2000);
                ClickRecompileButton(chromiumWebBrowser);

                // 7. Paste the original text
                PasteFromClipboard(chromiumWebBrowser, rawText);
            }
            catch (Exception)
            {
                MessageBox.Show("The document contains LaTeX commands that are not supported at the moment.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            finally
            {
                // Enable Main Menu
                MainWindow.EnableMainMenu(true);
            }
        }


        #region Filter
        /// <summary>
        /// Filters the text by removing all LaTeX commands which are not part of a predefined whitelist.
        /// </summary>
        /// <param name="rawText">The raw text.</param>
        /// <returns>The filtered text, which contains no LaTeX commands (except commands from the whitelist).</returns>
        private static string GetFilteredText(string rawText)
        {
            // Filters tags '\foo', '\foo{bar}', '\foo[42]{bar}', '\foo{bar}[42]', '\foo[13]{bar}[37]' and comments '% foo bar'
            var matches = Regex.Matches(rawText, PatternFilter);

            var latexCommands = matches.Cast<Match>().Select(m => m.Value).ToList();
            var latexCommandsToRemove = latexCommands.ToList(); // Deep Copy
            var filteredText = rawText;
            foreach (var latexCommand in latexCommands)
            {
                foreach (var word in _whitelist)
                {
                    // If the list of LaTeX commands contains a word from the whitelist,
                    // remove it from that list, so it will not be removed when filtering the text in the next step.
                    if (latexCommand.Contains(word))
                    {
                        latexCommandsToRemove.Remove(latexCommand);
                        break;
                    }
                }
            }

            // Remove all remaining LaTeX commands from the raw text
            foreach (var latexCommandToRemove in latexCommandsToRemove)
            {
                if (latexCommandToRemove.Length > 0)
                {
                    filteredText = filteredText.Replace(latexCommandToRemove, "");
                }
            }

            return filteredText;
        }
        #endregion


        #region Extract
        /// <summary>
        /// Extracts the sentences of the text and returns a set of sentences (no duplicates).
        /// It also extracts the plain text for each of the sentences without any LaTeX command.
        /// </summary>
        /// <param name="text">The text to extract the sentences from.</param>
        /// <returns>The set of sentences of the text without duplicates.</returns>
        private static HashSet<Sentence> ExtractSentencesFromText(string text)
        {
            // Annotation
            var annotation = new Annotation(text);
            _pipeline.annotate(annotation);

            // These are all the sentences in this document
            // a CoreMap is essentially a Map that uses class objects as keys and has values with custom types
            var sentences = annotation.get(typeof(CoreAnnotations.SentencesAnnotation));
            if (sentences == null)
            {
                return null;
            }

            // For each sentence create a new sentence object and add it to the dictionary if it is not already in there
            var dict = new Dictionary<string, Sentence>();
            foreach (Annotation sentence in sentences as ArrayList)
            {
                var s = new Sentence
                {
                    Text = sentence.toString(),
                    PlainText = Regex.Replace(sentence.toString(), PatternExtractPlaintext, @"$1"), // Extract plain text
                    Verbs = new List<string>(),
                    Nouns = new List<string>()
                };

                if (!dict.ContainsKey(sentence.toString()))
                {
                    dict.Add(sentence.toString(), s);
                }
            }

            return new HashSet<Sentence>(dict.Values);
        }
        #endregion


        #region Analyze
        /// <summary>
        /// Analyzes all the sentences.
        /// </summary>
        /// <param name="sentences">The sentences to analyze.</param>
        private static void AnalyzeSentences(HashSet<Sentence> sentences)
        {
            foreach (var sentence in sentences)
            {
                AnalyzeSentence(sentence);
            }
        }

        /// <summary>
        /// Analyzes a sentence by calculating the following five features per sentence.
        /// 1. Word Length
        /// 2. Vocabulary Complexity
        /// 3. Noun to Verb Ratio
        /// 4. Sentence Length
        /// 5. Sentence Structure Complexity
        /// 
        /// Additionally, it calculates the average for these five features and gets the corresponding color for that value from a color gradient.
        /// </summary>
        /// <param name="sentence">The sentence to analyze.</param>
        private static void AnalyzeSentence(Sentence sentence)
        {
            // Annotation
            var annotation = new Annotation(sentence.PlainText);
            _pipeline.annotate(annotation);

            // MIND: It is only one sentence because we analyze only one at a time!
            var sentences = annotation.get(typeof(CoreAnnotations.SentencesAnnotation)) as ArrayList;
            foreach (CoreMap coreMap in sentences)
            {
                // Get the tokens/words
                var tokens = coreMap.get(new CoreAnnotations.TokensAnnotation().getClass()) as ArrayList;

                var sentenceLength = 0;
                var vocabularyComplexity = 0;
                var totalAmountOfCharacters = 0;
                var minWordLength = int.MaxValue;
                var maxWordLength = int.MinValue;

                // Traversing the words in the current sentence
                foreach (CoreLabel token in tokens)
                {
                    // Get the word and part of speech from the token
                    var word = token.get(typeof(CoreAnnotations.TextAnnotation)).ToString();
                    var partOfSpeech = token.get(typeof(CoreAnnotations.PartOfSpeechAnnotation)).ToString();

                    // Do not process punctation marks!
                    if (word.Length == 1 && char.IsPunctuation(word.charAt(0)))
                    {
                        continue;
                    }

                    // Check if the word is a verb or noun
                    if (Verbs.Contains(partOfSpeech))
                    {
                        sentence.Verbs.Add(word);
                    }
                    else if (Nouns.Contains(partOfSpeech))
                    {
                        sentence.Nouns.Add(word);
                    }

                    // Check if the word is not in the dictionary (case insensitive)
                    if (!Settings.Dictionary.Contains(word, StringComparer.OrdinalIgnoreCase))
                    {
                        vocabularyComplexity++;
                    }


                    // Accumulate the word lengths and count the number of words
                    totalAmountOfCharacters += word.Length;
                    sentenceLength++;

                    // Determine the max/min length of all words
                    minWordLength = Math.Min(minWordLength, word.Length);
                    maxWordLength = Math.Max(maxWordLength, word.Length);
                }


                // Get the average word length and normalize to [0, 1]
                sentence.AverageWordLength = ((float)totalAmountOfCharacters / sentenceLength) / maxWordLength;

                // Get the vocabulary complexity and normalize to [0, 1]
                sentence.VocabularyComplexity = (float)vocabularyComplexity / sentenceLength;

                // Get the nound to verb ratio and normalize to [0, 1] 
                // (a higher value means more difficult to read)
                sentence.NounToVerbRatio = sentence.Nouns.Count > 0 ? (float)sentence.Nouns.Count / (sentence.Nouns.Count + sentence.Verbs.Count) : 0.0f;

                // Get the sentence length and normalize to [0, 1] using a predefined max sentence length
                // Everything above is hard to read and therefore 1
                sentence.SentenceLength = sentenceLength > MaxSentenceLength ? 1.0f : (float)sentenceLength / MaxSentenceLength;

                // Get the sentence structure complexity and normalize to [0, 1]
                var matches = Regex.Matches(sentence.PlainText, PatterSentenceStructureComplexity);
                sentence.SentenceStructureComplexity = (float)matches.Count / sentenceLength;

                // Calculate the average of the sentence using the weights from the settings/GUI
                sentence.AverageForSentence = Settings.AverageWordLengthWeight * sentence.AverageWordLength +
                                              Settings.VocabularyComplexityWeight * sentence.VocabularyComplexity + 
                                              Settings.NominalFormsWeight * sentence.NounToVerbRatio + 
                                              Settings.SentenceLengthWeight * sentence.SentenceLength + 
                                              Settings.SentenceStructureComplexityWeight * sentence.SentenceStructureComplexity;

                // Get the corresponding color from the color gradient for the average of the sentence
                sentence.Color = GetColor(sentence.AverageForSentence);
            }
        }
        #endregion


        #region Highlight
        /// <summary>
        /// Highlights each sentence in the calculated color of the average readability.
        /// </summary>
        /// <param name="sentences">The sentences to highlight.</param>
        /// <param name="rawText">The raw text to search in and highlight.</param>
        /// <returns>The highlighted text.</returns>
        private static string GetHighlightedText(HashSet<Sentence> sentences, string rawText)
        {
            var highlightedText = rawText;
            foreach (var sentence in sentences)
            {
                var r = sentence.Color.R / 255.0f;
                var g = sentence.Color.G / 255.0f;
                var b = sentence.Color.B / 255.0f;
                var colorString = string.Format("{{{0:0.0000} {1:0.0000} {2:0.0000}}}", r, g, b);

                // Inject LaTeX command to highlight the sentence
                var highlightedSentence = string.Format(@"\pdfmarkupcomment[markup=Highlight,color={0}]{{{1}}}{{{2}}}", colorString, sentence.Text, sentence.GenerateComment());
                highlightedText = highlightedText.Replace(sentence.Text, highlightedSentence);
            }

            highlightedText = highlightedText.Replace(Document, Package + Document);
            return highlightedText;
        }

        /// <summary>
        /// Calculates the corresponding color for a given value on a color gradient. The value needs to be in the range [0, 1].
        /// </summary>
        /// <param name="value">The value to calculate the color for.</param>
        /// <returns>The corresponding color.</returns>
        private static Color GetColor(float value)
        {
            int r, g, b;
            // Linear interpolation from blue (left) to white (center)
            if (value < 0.5)
            {
                r = (int)Math.Round(value * 2 * 255);
                g = r;
                b = 255;
            }
            // Linear interpolation from white (center) to red (right)
            else
            {
                r = 255;
                b = (int)Math.Round(((1 - value) * 2) * 255);
                g = b;
            }

            var color = Color.FromArgb(Alpha, r, g, b);
            return ConvertRgbaToRgb(color);
        }

        /// <summary>
        /// Converts from RGBA to RGB by interpolating between color and background color using alpha. 
        /// (1 - alpha) * RGB_background + alpha * RGBA_color, for the channels RGB and clamped to [0, 255].
        /// </summary>
        /// <param name="color">The highlight color to convert.</param>
        /// <returns>The converted color.</returns>
        private static Color ConvertRgbaToRgb(Color color)
        {
            var a = color.A / 255.0f;
            var r = Math.Min(Math.Max((1.0f - a) * BackgroundColor.R + a * color.R, 0), 255);
            var g = Math.Min(Math.Max((1.0f - a) * BackgroundColor.G + a * color.G, 0), 255);
            var b = Math.Min(Math.Max((1.0f - a) * BackgroundColor.B + a * color.B, 0), 255);
            return Color.FromArgb(255, (int) r, (int) g, (int) b); 
        }
        #endregion

        
        #region Javascript magic
        /// <summary>
        /// Presses the recompile button in the browser by using javascript.
        /// </summary>
        /// <param name="chromiumWebBrowser">The browser to use javascript with.</param>
        private static void ClickRecompileButton(ChromiumWebBrowser chromiumWebBrowser)
        {
            chromiumWebBrowser.ExecuteScriptAsync(Script);
        }
        #endregion
        

        #region Clipboard magic
        /// <summary>
        /// Copies the text from the browser to the clipboard.
        /// </summary>
        /// <param name="chromiumWebBrowser">The browser to copy from.</param>
        /// <returns>The copied text.</returns>
        private static string CopyToClipboard(ChromiumWebBrowser chromiumWebBrowser)
        {
            chromiumWebBrowser.Focus();
            chromiumWebBrowser.SelectAll();
            chromiumWebBrowser.Copy();
            Thread.Sleep(200);
            return GetTextFromClipboard();
        }

        /// <summary>
        /// Pastes the text from the clipboard to the browser.
        /// </summary>
        /// <param name="chromiumWebBrowser">The browser to copy from.</param>
        /// <param name="text">The text to paste.</param>
        private static void PasteFromClipboard(ChromiumWebBrowser chromiumWebBrowser, string text)
        {
            SetTextToClipboard(text);
            Thread.Sleep(200);
            chromiumWebBrowser.Focus();
            chromiumWebBrowser.SelectAll();
            chromiumWebBrowser.Paste();
        }

        /// <summary>
        /// Helper method, that recursively tries to get the text from the clipboard every 200ms, since it can be locked sometimes and cause an exception.
        /// </summary>
        /// <returns>The text from the clipboard.</returns>
        private static string GetTextFromClipboard()
        {
            string text;
            try
            {
                text = Clipboard.GetText(TextDataFormat.Text);
            }
            catch (Exception)
            {
                Thread.Sleep(200);
                text = GetTextFromClipboard();
            }
            return text;
        }

        /// <summary>
        /// Helper method, that recursively tries to set the text for the clipboard every 200ms, since it can be locked sometimes and cause an exception.
        /// </summary>
        /// <param name="text">The text for the clipboard.</param>
        private static void SetTextToClipboard(string text)
        {
            try
            {
                Clipboard.SetText(text);
            }
            catch (Exception)
            {
                Thread.Sleep(200);
                SetTextToClipboard(text);
            }
        }
        #endregion
    }
}