﻿using System.Collections.Generic;
using System.IO;
using System.Windows;
using Microsoft.Win32;
using ShareReadability.Container;
using ShareReadability.Util;

namespace ShareReadability
{
    /// @file SettingsWindow.xaml.cs
    /// <inheritdoc cref="Window" />
    /// <summary>
    /// Interaction logic for SettingsWindow.xaml
    /// </summary>
    public partial class SettingsWindow : Window
    {
        public Settings Settings { get; set; }
        private readonly List<CustomSlider> _sliders;
        private string _dictionaryPath;
        private double _sum;

        /// <inheritdoc />
        /// <summary>
        /// Constructor of the settings window.
        /// </summary>
        public SettingsWindow()
        {
            InitializeComponent();
            _sliders = new List<CustomSlider> {SliderAwl, SliderSl, SliderVc, SliderNf, SliderSsc};
        }

        /// <summary>
        /// Initializes the sliders.
        /// </summary>
        public void InitializeSlider()
        {
            SetSliderInActive();
            SliderAwl.Value = Settings.AverageWordLengthWeight;
            SliderSl.Value = Settings.SentenceLengthWeight;
            SliderVc.Value = Settings.VocabularyComplexityWeight;
            SliderNf.Value = Settings.NominalFormsWeight;
            SliderSsc.Value = Settings.SentenceStructureComplexityWeight;
            SetSliderActive();
            _dictionaryPath = Settings.DictionaryPath;
            TextBlockDictionaryPath.Text = _dictionaryPath;
        }

        /// <summary>
        /// Closes the settings window and sets the values.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The event.</param>
        private void ButtonOk_Click(object sender, RoutedEventArgs e)
        {
            // Set values from GUI
            Settings.DictionaryPath = _dictionaryPath;
            Settings.Dictionary = new List<string>(File.ReadAllLines(Settings.DictionaryPath));

            Settings.AverageWordLengthWeight = (float)SliderAwl.Value;
            Settings.SentenceLengthWeight = (float)SliderSl.Value;
            Settings.VocabularyComplexityWeight = (float)SliderVc.Value;
            Settings.NominalFormsWeight = (float)SliderNf.Value;
            Settings.SentenceStructureComplexityWeight = (float)SliderSsc.Value;

            Close();
        }

        /// <summary>
        /// Closes the settings window without passing the set values.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The event.</param>
        private void ButtonCancel_Click(object sender, RoutedEventArgs e)
        {
            Close();
        }

        /// <summary>
        /// Sets all sliders active.
        /// </summary>
        private void SetSliderActive()
        {
            foreach (var slider in _sliders)
            {
                slider.DontRunHandler = false;
            }
        }

        /// <summary>
        /// Sets all sliders inactive.
        /// </summary>
        private void SetSliderInActive()
        {
            foreach (var slider in _sliders)
            {
                slider.DontRunHandler = true;
            }
        }

        /// <summary>
        /// Normalizes all slider values with their total sum, so they sum up to 1.0.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The event.</param>
        private void Slider_OnValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            var currentSlider = sender as CustomSlider;
            if (currentSlider != null && currentSlider.DontRunHandler) return;
            SetSliderInActive();

            _sum = 0.0;
            foreach (var customSlider in _sliders)
            {
                _sum += customSlider.Value;
            }

            foreach (var customSlider in _sliders)
            {
                customSlider.Value /= _sum;
            }

            SetSliderActive();
        }

        /// <summary>
        /// Loads a new dictionary, if the actions is not canceled.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The event.</param>
        private void ButtonDictionary_OnClick(object sender, RoutedEventArgs e)
        {
            var openFileDialog = new OpenFileDialog
            {
                Multiselect = false,
                Filter = @"Text files(*.txt)|*.txt|All files(*.*)|*.*"
            };

            var combinedPath = Path.Combine(Directory.GetCurrentDirectory(), "data/dictionary");
            openFileDialog.InitialDirectory = Path.GetFullPath(combinedPath);
            
            if (openFileDialog.ShowDialog() == true)
            {
                var filename = Path.GetFileName(openFileDialog.FileName);
                TextBlockDictionaryPath.Text = filename;
                _dictionaryPath = openFileDialog.FileName;
            }
        }
    }
}